import datetime
import importlib
import inspect
import json
import os
import pytz
import requests
import sys
import time
from pprint import pprint

from django.conf import settings
from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.core.mail import mail_admins
from django.core.management.base import BaseCommand, CommandError
from django.db.models import F
from django.utils import timezone

from saleboxdjango.lib.common import api_auth, dictfetchall, update_natural_sort
from saleboxdjango.lib.translation import get_translations
from saleboxdjango.models import *


class Command(BaseCommand):
    def handle(self, *args, **options):
        now = time.time()

        # bail out in sync already running
        sync_start = int(self.timer_get("saleboxsync_sync_start"))
        if (time.time() - sync_start) < 120:
            print("sync in progress: bailing out!")
            return

        # step #1: if a user needs to be converted to a member, queue it
        self.timer_set("saleboxsync_sync_start", time.time())
        user_equals_member = settings.SALEBOX["MEMBER"]["USER_EQUALS_MEMBER"]
        if user_equals_member != "MANUAL":
            # 'ALWAYS': every website user should be a salebox member
            if user_equals_member == "ALWAYS":
                user_ids = (
                    get_user_model()
                    .objects.filter(salebox_member_id__isnull=True)
                    .filter(last_login__isnull=False)
                    .values_list("id", flat=True)
                )

            # 'PURCHASE': every website user that makes a purchase should
            # be a salebox member
            if user_equals_member == "PURCHASE":
                user_ids = (
                    CheckoutStore.objects.filter(status__lte=30)
                    .filter(user__isnull=False)
                    .values_list("user", flat=True)
                )
                if len(user_ids) > 0:
                    user_ids = (
                        get_user_model()
                        .objects.filter(id__in=list(set(user_ids)))
                        .filter(salebox_member_id__isnull=True)
                        .values_list("id", flat=True)
                    )

            # give those users a salebox_member_id and details to sync
            if len(user_ids) > 0:
                users = get_user_model().objects.filter(id__in=user_ids)
                for u in users:
                    u.set_salebox_member_sync("email", u.email)
                    # potential TODO:
                    # have a list of key mappings in the config
                    # which automatically get added to the sync
                    # JSON at this point

        # step #2: push users with member updates
        self.timer_set("saleboxsync_sync_start", time.time())
        self.push_members()

        # step #3: pull data
        # members must be up-to-date before POSTing transactions
        pull_start = int(self.timer_get("saleboxsync_pull_start"))
        # pull_start = 0
        if (time.time() - pull_start) > 599:
            self.pull_loop()

        # step 4: POST transactions
        transactions_pushed = self.push_transactions()
        if transactions_pushed:
            # pull again to get the event emails
            time.sleep(2)
            self.pull_loop()

        # step 5: run event handler
        self.timer_set("saleboxsync_sync_start", time.time())
        if "EVENT_HANDLER" in settings.SALEBOX:
            sys.path.insert(0, settings.SALEBOX["EVENT_HANDLER"]["APP_FOLDER"])
            import_path = settings.SALEBOX["EVENT_HANDLER"]["CLASS"]
            path, name = import_path.rsplit(".", 1)
            c = getattr(importlib.import_module(path), name)
            c()

        # step 6: send django-mail-queue emails
        try:
            self.timer_set("saleboxsync_sync_start", time.time())
            from mailqueue.models import MailerMessage

            MailerMessage.objects.send_queued(limit=30)
        except:
            pass

        # step 7: update stock levels
        # self.pull_inventory()

        # step 8: pull latest member transaction history
        self.pull_transaction_history()

        # step 9: send analytics
        # self.push_footfall()

        # step 10: check for timed-out checkout stores
        # CheckoutStore.objects.apply_timeout()

        # finished: reset the clock
        self.timer_set("saleboxsync_sync_start", 0.0)
        print("Finished in %.3fs" % (time.time() - now))

    def generate_sequential_id(self, store):
        tz = pytz.timezone(settings.TIME_ZONE)
        payment_recvd = store.payment_received.astimezone(tz)

        count = (
            CheckoutStore.objects.filter(status__in=[30, 31])
            .exclude(id=store.id)
            .filter(payment_received__lt=payment_recvd)
            .filter(payment_received__year=payment_recvd.year)
            .filter(payment_received__month=payment_recvd.month)
        )
        if settings.SALEBOX["CHECKOUT"]["SEQUENTIAL_TRANSACTION_ID"]["RATE"] == "daily":
            count = count.filter(payment_received__day=payment_recvd.day)

        # fetch format
        prefix = settings.SALEBOX["CHECKOUT"]["SEQUENTIAL_TRANSACTION_ID"]["PREFIX"]
        digits = settings.SALEBOX["CHECKOUT"]["SEQUENTIAL_TRANSACTION_ID"]["DIGITS"]

        # replace characters in prefix
        prefix = prefix.replace("{yyyy}", str(payment_recvd.year))
        prefix = prefix.replace("{yy}", str(payment_recvd.year)[-2:])
        prefix = prefix.replace("{mm}", ("0" + str(payment_recvd.month))[-2:])
        prefix = prefix.replace("{dd}", ("0" + str(payment_recvd.day))[-2:])

        count = count.count() + 1
        return f"{prefix}{count:0{digits}d}"

    def pull(self):
        post = api_auth()

        # populate last updates
        sync_from_dict = self.pull_get_sync_from_dict()
        for code in sync_from_dict:
            post["lu_%s" % code] = sync_from_dict[code]

        print()
        print("Attempting sync with the following parameters:")
        pprint(sync_from_dict)
        print()

        # do request
        url = "%s/api/pos/v2/sync" % settings.SALEBOX["API"]["URL"]
        try:
            r = requests.post(url, data=post)
            have_response = True
        except:
            self.send_admin_email("Could not connect to Salebox POSv2 API")
            have_response = False

        try:
            if have_response:
                response = r.json()
                # pprint(response)

                # display debug meta
                print("Output received ('data' attribute truncated to len(data)):")
                response_meta = r.json()

                # exit if connection error
                if response_meta["status"] != "OK":
                    print("Connection error: %s" % response_meta["status"])
                    return

                # display meta object
                for code in response_meta["sync"]:
                    code["data"] = len(code["data"])
                pprint(response_meta)
                print()
                print("Updating:")

                # save data
                for i, value in enumerate(response["sync"]):
                    if value["code"] == "attribute":
                        self.pull_model_attribute(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "attribute_item":
                        self.pull_model_attribute_item(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "country":
                        self.pull_model_country(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "country_state":
                        self.pull_model_country_state(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "discount_group":
                        self.pull_model_discount_group(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "discount_ruleset":
                        self.pull_model_discount_ruleset(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "event":
                        self.pull_model_event(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "i18n":
                        self.pull_model_translation(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "member":
                        self.pull_model_member(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "member_group":
                        self.pull_model_member_group(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "product":
                        self.pull_model_product(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "product_category":
                        self.pull_model_product_category(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "product_variant":
                        self.pull_model_product_variant(
                            value["data"], value["lu"], sync_from_dict
                        )

                    elif value["code"] == "product_variant_image":
                        self.pull_model_product_variant_image(
                            value["data"], value["lu"], sync_from_dict
                        )

                    else:
                        print("Error: %s" % value["code"])

                return response["resync_now"]
                # return False
        except:
            self.send_admin_email()
            return False

    def pull_get_sync_from_dict(self):
        lus = LastUpdate.objects.all()

        # populate dict
        lu = {}
        for l in lus:
            lu[l.code] = l.value

        # fill in the gaps
        for code in [
            "attribute",
            "attribute_item",
            "country",
            "country_state",
            # 'discount_group',
            # 'discount_ruleset',
            "event",
            "i18n",
            "member",
            "member_group",
            # 'product',
            # 'product_category',
            # 'product_variant',
            "product_variant_image",
        ]:
            if code not in lu:
                LastUpdate(code=code, value=0.0).save()
                lu[code] = 0.0

        return lu

    def pull_image(self, id, urlpath, dir):
        self.timer_set("saleboxsync_sync_start", time.time())
        self.timer_set("saleboxsync_pull_start", time.time())

        # ensure target dir exists
        target_dir = "%s/salebox/%s" % (settings.MEDIA_ROOT, dir)
        try:
            os.makedirs(target_dir)
        except:
            pass

        # fetch image
        try:
            url = "%s/%s" % (settings.SALEBOX["API"]["URL"], urlpath)
            suffix = urlpath.split(".")[-1]
            filename = "%s.%s.%s" % (id, urlpath.split("/")[-1][0:6], suffix)
            target = "%s/%s" % (target_dir, filename)
            r = requests.get(url)
            if r.status_code == 200:
                open(target, "wb").write(r.content)
                return filename, True
        except:
            self.send_admin_email()

        return None, False

    def pull_images(self):
        # # product category
        # imgs = ProductCategory \
        #         .objects \
        #         .filter(active_flag=True) \
        #         .filter(image__isnull=False) \
        #         .filter(local_image__isnull=True)
        # for img in imgs:
        #     path, success = self.pull_image(img.id, img.image[1:], 'pospc')
        #     if success:
        #         print('.', end='')
        #         img.local_image = path
        #         img.save()

        # # product
        # imgs = Product \
        #         .objects \
        #         .filter(active_flag=True) \
        #         .filter(image__isnull=False) \
        #         .filter(local_image__isnull=True)
        # for img in imgs:
        #     path, success = self.pull_image(img.id, img.image[1:], 'posp')
        #     if success:
        #         print('.', end='')
        #         img.local_image = path
        #         img.save()

        # # product variant
        # imgs = ProductVariant \
        #         .objects \
        #         .filter(active_flag=True) \
        #         .filter(image__isnull=False) \
        #         .filter(local_image__isnull=True)
        # for img in imgs:
        #     path, success = self.pull_image(img.id, img.image[1:], 'pospv')
        #     if success:
        #         print('.', end='')
        #         img.local_image = path
        #         img.save()

        # product variant image
        imgs = (
            ProductVariantImage.objects.exclude(img__isnull=True)
            .filter(local_img__isnull=True)
            .filter(active_flag=True)
        )
        for img in imgs:
            path, success = self.pull_image(img.id, img.img, "pvi")
            if success:
                print(".", end="")
                img.local_img = path
                img.save()

    # def pull_inventory(self):
    #     url = "%s/api/pos/v2/inventory" % settings.SALEBOX["API"]["URL"]
    #     post = api_auth()
    #     inventory = []
    #
    #     # attempt sync 'all'
    #     sync_all = int(self.timer_get("saleboxsync_inventory_all"))
    #     # sync_all = 0
    #     if (time.time() - sync_all) > (60 * 60 * 24):
    #         print("Inventory sync all")
    #         post["request"] = "all"
    #
    #         # update checked-out stock
    #         # this shouldn't be needed, but a once-a-day-tidy up won't hurt
    #         CheckoutStore.objects.update_checked_out_stock()
    #     else:
    #         # attempt sync 'recent'
    #         sync_recent = int(self.timer_get("saleboxsync_inventory_recent"))
    #         # sync_recent = 999
    #         if (time.time() - sync_recent) > (60 * 5):
    #             print("Inventory sync recent")
    #             post["request"] = "recent"
    #             post["time_offset"] = (
    #                 60 * 30
    #             )  # everything that changed in the last 30 minutes
    #         else:
    #             # attempt sync 'lowstock'
    #             variant_ids = (
    #                 ProductVariant.objects.filter(active_flag=True)
    #                 .filter(product__active_flag=True)
    #                 .filter(product__category__active_flag=True)
    #                 .filter(available_on_ecom=True)
    #                 .filter(stock_total__gt=0)
    #                 .filter(stock_total__lte=F("ecommerce_low_stock_threshold"))
    #                 .values_list("id", flat=True)
    #             )
    #             if len(variant_ids) > 0:
    #                 print("Inventory sync low-stock")
    #                 post["request"] = "lowstock"
    #                 post["variant_ids"] = ",".join([str(i) for i in variant_ids])
    #
    #     # bail if nothing to fetch
    #     if "request" not in post:
    #         return
    #
    #     # fetch from server
    #     try:
    #         r = requests.post(url, data=post)
    #         if r.status_code == 200:
    #             result = r.json()
    #             if result["status"] == "OK":
    #                 if "inventory" in result:
    #                     inventory = result["inventory"]
    #                 else:
    #                     self.send_admin_email(
    #                         "API response recieved but inventory missing (request type: %s)"
    #                         % post["request"]
    #                     )
    #             else:
    #                 self.send_admin_email(
    #                     "API response recieved but status not OK (request type: %s)"
    #                     % post["request"]
    #                 )
    #         else:
    #             self.send_admin_email(
    #                 "API response status code: %s (request type: %s)"
    #                 % (r.status_code, post["request"])
    #             )
    #     except:
    #         self.send_admin_email(
    #             "Unknown error connecting to Salebox POSv2 API (request type: %s)"
    #             % post["request"]
    #         )
    #
    #     # apply results
    #     if len(inventory) > 0:
    #         for i in inventory:
    #             variant_id = i[0]
    #             stock_count = i[1]
    #             pv = ProductVariant.objects.filter(id=variant_id).first()
    #             if pv is not None:
    #                 if pv.stock_count != stock_count:
    #                     pv.stock_count = stock_count
    #                     pv.save()
    #
    #         # set timers
    #         if post["request"] == "all":
    #             self.timer_set("saleboxsync_inventory_all", time.time())
    #             self.timer_set("saleboxsync_inventory_recent", time.time())
    #         elif post["request"] == "recent":
    #             self.timer_set("saleboxsync_inventory_recent", time.time())

    def pull_loop(self):
        while True:
            self.timer_set("saleboxsync_sync_start", time.time())
            self.timer_set("saleboxsync_pull_start", time.time())
            pull_result = self.pull()
            if pull_result:
                print()
                print("sleeping...")
                print()
                time.sleep(5)
            else:
                break

        # pull images
        print()
        print("Downloading images...")
        self.pull_images()
        print()

    def pull_model_attribute(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = Attribute.objects.get_or_create(id=d["id"])
                o.code = d["code"]
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "attribute", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x Attribute" % len(data))
        except:
            self.send_admin_email()

    def pull_model_attribute_item(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = AttributeItem.objects.get_or_create(id=d["id"])
                o.attribute = Attribute.objects.get(id=d["attribute"])
                o.slug = d["slug"]
                o.value = d["value"]
                o.extras = d["extras"]
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "attribute_item", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x AttributeItem" % len(data))
        except:
            self.send_admin_email()

    def pull_model_country(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = Country.objects.get_or_create(id=d["id"])
                o.code = d["code"]
                o.default = d["default"]
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "country", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x Country" % len(data))
        except:
            self.send_admin_email()

    def pull_model_country_state(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = CountryState.objects.get_or_create(id=d["id"])
                o.country = Country.objects.get(id=d["country"])
                o.code = d["code"]
                o.full_code = d["full_code"]
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "country_state", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x CountryState" % len(data))
        except:
            self.send_admin_email()

    def pull_model_discount_group(self, data, api_lu, sync_from_dict):
        # for d in data
        #
        #

        # update sync_from
        """
        self.pull_set_sync_from_dict(
            'discount_group',
            len(data) < 100,
            sync_from_dict,
            api_lu
        )
        """

    def pull_model_discount_ruleset(self, data, api_lu, sync_from_dict):
        # for d in data
        #
        #

        # update sync_from
        """
        self.pull_set_sync_from_dict(
            'discount_ruleset',
            len(data) < 100,
            sync_from_dict,
            api_lu
        )
        """

    def pull_model_event(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = Event.objects.get_or_create(id=d["id"])
                o.event = d["event"]
                o.salebox_member_id = d["salebox_member_id"]
                o.transaction_guid = d["transaction_guid"]
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "event", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x Event" % len(data))
        except:
            self.send_admin_email()

    def pull_model_member(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                # retrieve lookups
                group = MemberGroup.objects.get(id=d["group"])
                parent = Member.objects.get_or_create(id=d["id"])
                try:
                    gwc = MemberGroup.objects.get(id=d["group_when_created"])
                except:
                    gwc = None

                # get parent
                parent = None
                if d["parent"] is not None:
                    parent, created = Member.objects.get_or_create(id=d["parent"])

                # country / states
                country = None
                country_state = None
                if d["country"] is not None:
                    country = Country.objects.get(id=d["country"])
                if d["country_state"] is not None:
                    country_state = CountryState.objects.get(id=d["country_state"])

                # create object
                o, created = Member.objects.get_or_create(id=d["id"])
                o.group = group
                o.parent = parent
                o.group_when_created = gwc
                o.country = country
                o.country_state = country_state

                # update
                for a in [
                    "active_flag",
                    "address_1",
                    "address_2",
                    "address_3",
                    "address_4",
                    "address_5",
                    "boolean_1",
                    "boolean_2",
                    "boolean_3",
                    "boolean_4",
                    "boolean_5",
                    "boolean_6",
                    "date_of_birth",
                    "email",
                    "gender",
                    "guid",
                    "id_card",
                    "join_date",
                    "name_first",
                    "name_last",
                    "phone_1",
                    "phone_2",
                    "postcode",
                    "salebox_member_id",
                    "status",
                    "string_1",
                    "string_2",
                    "string_3",
                    "string_4",
                    "string_5",
                    "string_6",
                    "string_7",
                    "string_8",
                    "string_9",
                    "string_10",
                    "string_11",
                    "string_12",
                    "string_13",
                    "string_14",
                    "title",
                ]:
                    setattr(o, a, d[a])

                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "member", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x Member" % len(data))
        except:
            self.send_admin_email()

    def pull_model_member_group(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                o, created = MemberGroup.objects.get_or_create(id=d["id"])
                o.name = d["name"]
                o.flat_discount_percentage = d["flat_discount_percentage"]
                o.can_be_parent = d["can_be_parent"]
                o.default_group = d["default_group"]
                o.active_flag = d["active_flag"]
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "member_group", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x MemberGroup" % len(data))
        except:
            self.send_admin_email()

    def pull_model_product(self, data, api_lu, sync_from_dict):
        def sync_attributes(product, attribute_number, id_list):
            attributes_changed = False
            attribute_m2m = getattr(product, "attribute_%s" % attribute_number)

            # remove
            for a in attribute_m2m.all():
                if a.id not in id_list:
                    attribute_m2m.remove(a)
                    attributes_changed = True

            # add
            for id in id_list:
                ai = AttributeItem.objects.get(id=id)
                attribute_m2m.add(ai)
                attributes_changed = True

            # save
            if attributes_changed:
                product.save()

        try:
            for d in data:
                # create object
                o, created = Product.objects.get_or_create(id=d["id"])
                o.category = ProductCategory.objects.get(id=d["category"])

                # do we need to reset the local_image flag?
                if not created and o.image != d["image"]:
                    o.local_image = None

                # update
                for a in [
                    "active_flag",
                    "bestseller_rank",
                    "boolean_1",
                    "boolean_2",
                    "boolean_3",
                    "boolean_4",
                    "boolean_5",
                    "boolean_6",
                    "boolean_7",
                    "boolean_8",
                    "boolean_9",
                    "boolean_10",
                    "i18n",
                    "image",
                    "inventory_type",
                    "name",
                    "season",
                    "sold_by",
                    "slug",
                    "string_1",
                    "string_2",
                    "string_3",
                    "string_4",
                    "string_5",
                    "string_6",
                    "string_7",
                    "string_8",
                    "string_9",
                    "string_10",
                    "text_1",
                    "text_2",
                    "text_3",
                    "text_4",
                    "text_5",
                    "text_6",
                    "text_7",
                    "text_8",
                    "text_9",
                    "text_10",
                    "vat_applicable",
                ]:
                    setattr(o, a, d[a])

                o.save()

                # sync attributes
                for i in range(1, 11):
                    sync_attributes(o, i, d["attribute_%s" % i])

            # update sync_from
            self.pull_set_sync_from_dict(
                "product", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x Product" % len(data))
        except:
            self.send_admin_email()

    def pull_model_product_category(self, data, api_lu, sync_from_dict):
        try:
            # pass #1: save with parent = None to avoid mptt errors
            for d in data:
                o, created = ProductCategory.objects.get_or_create(id=d["id"])

                # do we need to reset the local_image flag?
                if not created and o.image != d["image"]:
                    o.local_image = None

                o.parent = None
                o.active_flag = d["active_flag"]
                o.boolean_1 = d["boolean_1"]
                o.boolean_2 = d["boolean_2"]
                o.boolean_3 = d["boolean_3"]
                o.boolean_4 = d["boolean_4"]
                o.i18n = d["i18n"]
                o.image = d["image"]
                o.name = d["name"]
                o.seasonal_flag = d["seasonal_flag"]
                o.slug = d["slug"]
                o.short_name = d["short_name"]
                setattr(o, o._mptt_meta.left_attr, d["mptt_left"])
                setattr(o, o._mptt_meta.level_attr, d["mptt_level"])
                setattr(o, o._mptt_meta.parent_attr, None)
                setattr(o, o._mptt_meta.right_attr, d["mptt_right"])
                setattr(o, o._mptt_meta.tree_id_attr, d["mptt_tree_id"])
                o.save()

            # pass #2: re-save the categories with parents (where applicable)
            for d in data:
                parent = None
                if d["parent"] is not None:
                    parent = ProductCategory.objects.get(id=d["parent"])

                o = ProductCategory.objects.get(id=d["id"])
                o.parent = parent
                setattr(o, o._mptt_meta.parent_attr, parent)
                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "product_category", True, sync_from_dict, api_lu
            )

            print("%s x ProductCategory" % len(data))
        except:
            self.send_admin_email()

    def pull_model_product_variant(self, data, api_lu, sync_from_dict):
        def sync_attributes(variant, attribute_number, id_list):
            attributes_changed = False
            attribute_m2m = getattr(variant, "attribute_%s" % attribute_number)

            # remove
            for a in attribute_m2m.all():
                if a.id not in id_list:
                    attribute_m2m.remove(a)
                    attributes_changed = True

            # add
            for id in id_list:
                ai = AttributeItem.objects.get(id=id)
                attribute_m2m.add(ai)
                attributes_changed = True

            # save
            if attributes_changed:
                variant.save()

        try:
            for d in data:
                # create object
                o, created = ProductVariant.objects.get_or_create(id=d["id"])
                o.product = Product.objects.get(id=d["product"])

                # do we need to reset the local_image flag?
                if not created and o.image != d["image"]:
                    o.local_image = None

                # update
                for a in [
                    "active_flag",
                    "available_to_order",
                    "available_on_pos",
                    "available_on_ecom",
                    "bestseller_rank",
                    "barcode",
                    "bo_name",
                    "boolean_1",
                    "boolean_2",
                    "boolean_3",
                    "boolean_4",
                    "boolean_5",
                    "color",
                    "date_1",
                    "date_2",
                    "ecommerce_description",
                    "ecommerce_low_stock_threshold",
                    "i18n",
                    "image",
                    "int_1",
                    "int_2",
                    "int_3",
                    "int_4",
                    "int_5",
                    "loyalty_points",
                    "member_discount_applicable",
                    "name",
                    "plu",
                    "preorder_flag",
                    "price",
                    "sale_percent",
                    "sale_price",
                    "shelf_expiry_type",
                    "shelf_life_days",
                    "shipping_weight",
                    "size",
                    "size_order",
                    "size_uom",
                    "sku",
                    "slug",
                    "string_1",
                    "string_2",
                    "string_3",
                    "string_4",
                    "string_5",
                    "warehouse_location",
                ]:
                    setattr(o, a, d[a])

                o.save()

                # sync attributes
                for i in range(1, 11):
                    sync_attributes(o, i, d["attribute_%s" % i])

            # update sync_from
            self.pull_set_sync_from_dict(
                "product_variant", len(data) < 100, sync_from_dict, api_lu
            )

            # update the natural sort value
            if len(data) > 0:
                update_natural_sort(ProductVariant, "name", "name_sorted")

            print("%s x ProductVariant" % len(data))
        except:
            self.send_admin_email()

    def pull_model_product_variant_image(self, data, api_lu, sync_from_dict):
        try:
            for d in data:
                # create object
                o, created = ProductVariantImage.objects.get_or_create(
                    id=d["id"], variant=ProductVariant.objects.get(id=d["variant"])
                )

                # update
                for a in [
                    "active_flag",
                    "description",
                    "img",
                    "img_height",
                    "img_width",
                    "order",
                    "title",
                ]:
                    setattr(o, a, d[a])

                o.save()

            # update sync_from
            self.pull_set_sync_from_dict(
                "product_variant_image", len(data) < 100, sync_from_dict, api_lu
            )

            print("%s x ProductVariantImage" % len(data))
        except:
            self.send_admin_email()

    def pull_model_translation(self, data, api_lu, sync_from_dict):
        try:
            if "translations" in data:
                for language_code in data["translations"]:
                    self.pull_model_translation_recurse(
                        language_code, data["translations"][language_code]
                    )

            # update sync_from
            self.pull_set_sync_from_dict("i18n", True, sync_from_dict, api_lu)

            # rebuild cache
            get_translations(rebuild=True)

            print("%s x Translation" % len(data))
        except:
            self.send_admin_email()

    def pull_model_translation_recurse(self, language_code, data, prefix=""):
        for d in data:
            if isinstance(data[d], str):
                t, created = Translation.objects.get_or_create(
                    language_code=language_code, key="%s%s" % (prefix, d)
                )
                t.value = data[d]
                t.save()
            else:
                self.pull_model_translation_recurse(language_code, data[d], "%s." % d)

    def pull_set_sync_from_dict(self, code, increment, sync_from_dict, api_lu):
        if increment:
            if timezone.now().timestamp() - api_lu > 20:
                if api_lu == 0:
                    api_lu = 1.0
                else:
                    api_lu += 0.00001

        if api_lu > sync_from_dict[code]:
            lu = LastUpdate.objects.get(code=code)
            lu.value = float(api_lu)
            lu.save()
            cache.clear()

    def pull_transaction_history(self):
        members = Member.objects.filter(
            salebox_transactionhistory_request_dt__isnull=False
        ).order_by("salebox_transactionhistory_request_dt")[0:1000]

        for m in members:
            self.timer_set("saleboxsync_sync_start", time.time())
            url = "%s/api/pos/v2/member-transaction-history" % (
                settings.SALEBOX["API"]["URL"],
            )
            post = api_auth()
            post["salebox_member_id"] = str(m.salebox_member_id)

            # fetch from server
            try:
                r = requests.post(url, data=post)
                result = r.json()
                if "transactions" in result:
                    m.transactionhistory_update_data(result["transactions"])
            except:
                self.send_admin_email("Could not connect to Salebox POSv2 API")

    # def push_footfall(self):
    #     if (
    #         settings.SALEBOX["ANALYTICS"]["SEND"]
    #         and "postgres" in settings.DATABASES["default"]["ENGINE"]
    #     ):
    #         sql = """
    #             SELECT          COUNT(*) AS visitor_count
    #                             ,DATE_PART('hour', first_seen AT TIME ZONE 'UTC') AS hour
    #                             ,DATE_PART('day', first_seen AT TIME ZONE 'UTC') AS day
    #                             ,DATE_PART('month', first_seen AT TIME ZONE 'UTC') AS month
    #                             ,DATE_PART('year', first_seen AT TIME ZONE 'UTC') AS year
    #             FROM            saleboxdjango_analytic
    #             WHERE           first_seen >= TIMESTAMP 'yesterday' AT TIME ZONE 'UTC'
    #             AND             ua_is_bot = false
    #             GROUP BY        hour, day, month, year;
    #         """
    #
    #         # create list of dicts
    #         visitors = dictfetchall(sql)
    #         for v in visitors:
    #             for key in v.keys():
    #                 v[key] = int(v[key])
    #
    #         # send
    #         data = api_auth()
    #         data["visitors"] = json.dumps({"visitors": visitors})
    #         url = "%s/api/pos/v2/ecommerce-visitors" % settings.SALEBOX["API"]["URL"]
    #         try:
    #             r = requests.post(url, data=data)
    #         except:
    #             self.send_admin_email()

    def push_member(self, user):
        self.timer_set("saleboxsync_sync_start", time.time())

        # define a salebox_member_id if one doesn't exist
        if user.salebox_member_id is None:
            user.create_salebox_member_id()

        # build post
        post = api_auth()
        post["salebox_member_id"] = user.salebox_member_id
        post["salebox_member_sync"] = json.dumps(
            user.salebox_member_sync, ensure_ascii=False
        )

        # do request
        url = "%s/api/pos/v2/member-update" % settings.SALEBOX["API"]["URL"]
        try:
            r = requests.post(url, data=post)
            if r.status_code == 200:
                user.salebox_member_sync = None
                user.save()
                self.timer_set("saleboxsync_pull_start", 0.0)
        except:
            self.send_admin_email("Failed to POST member data")

    def push_members(self):
        if hasattr(get_user_model(), "salebox_member_sync"):
            users = get_user_model().objects.filter(salebox_member_sync__isnull=False)
            for user in users:
                self.push_member(user)
        else:
            print("SaleboxUser class not in use")

    def push_transaction(self, store):
        # build POST
        data = api_auth()
        data["transaction"] = json.dumps(
            {
                "basket": self.transaction_basket(store),
                "customer": self.transaction_customer(store),
                "extras": self.transaction_extras(store),
                "invoice": self.transaction_invoice(store),
                "manual_discount": None,
                "member": self.transaction_member(store),
                "messages": store.data.get("messages", {}),
                "meta": self.transaction_meta(store),
                "payment": self.transaction_payment(store),
                "shipping": self.transaction_shipping(store),
                "stored_value_load": None,
                "total": self.transaction_total(store),
            },
            ensure_ascii=False,
        )

        # send
        url = "%s/api/pos/v2/transaction/create" % settings.SALEBOX["API"]["URL"]
        try:
            r = requests.post(url, data=data)
            have_response = True
        except:
            self.send_admin_email()
            have_response = False

        # check response
        if have_response:
            if r.json()["status"] == "OK":
                store.status = 31
                store.save()

    def push_transactions(self):
        transactions_pushed = False

        css = CheckoutStore.objects.filter(status=30)
        for cs in css:
            self.timer_set("saleboxsync_sync_start", time.time())

            # ensure this user has a corresponding salebox_member
            user_equals_member = settings.SALEBOX["MEMBER"]["USER_EQUALS_MEMBER"]
            if user_equals_member != "MANUAL" and cs.user is not None:
                user = get_user_model().objects.get(id=cs.user)
                member = Member.objects.filter(
                    salebox_member_id=user.salebox_member_id
                ).first()
                if member is None:
                    # skip posting this transaction for now...
                    # get on the next time round
                    if user.salebox_member_sync is None:
                        user.set_salebox_member_sync("email", user.email)
                    continue

            # push transaction
            self.push_transaction(cs)
            transactions_pushed = True

        return transactions_pushed

    def send_admin_email(self, extra=None):
        caller = inspect.currentframe().f_back
        msg = [
            "saleboxsync error occurred:",
            "",
            "         Website: %s" % settings.ALLOWED_HOSTS[0],
            "      Backoffice: %s" % settings.SALEBOX["API"]["URL"],
            "  Error function: %s" % caller.f_code.co_name,
            "    Error lineno: %s" % caller.f_lineno,
        ]

        if extra is not None:
            msg.append("           Notes: %s" % extra)

        if settings.DEBUG:
            print(
                "--------------------------------------------------------------------------"
            )
            print("\n".join(msg))
            print(
                "--------------------------------------------------------------------------"
            )
        else:
            mail_admins(
                "saleboxsync error: %s" % settings.ALLOWED_HOSTS[0], "\n".join(msg)
            )

    def timer_get(self, code, default_value=0.0):
        try:
            return LastUpdate.objects.get(code=code).value
        except:
            LastUpdate(code=code, value=default_value).save()
            return default_value

    def timer_set(self, code, value):
        o = LastUpdate.objects.filter(code=code).first()
        if o is None:
            o = LastUpdate(code=code)
        o.value = value
        o.save()

    def transaction_basket(self, store):
        basket = []
        for b in store.data["basket"]["items"]:
            v = b["variant"]

            basket.append(
                {
                    "discount_ruleset": None,
                    "qty_loyalty": v.get("qty_loyalty", None),
                    "price_original": v["qty_price"],
                    "price_modified": v["qty_sale_price"],
                    "product_weight": None,
                    "quantity": b["qty"],
                    "shipping_weight": v["shipping_weight"],
                    "unit_price": v["price"],
                    "variant_id": v["id"],
                    "weight": b.get("weight", 0),
                }
            )

        return basket

    def transaction_customer(self, store):
        # w/c 30 september 2019 uncomment this line...
        # return store.data['customer']

        # ... send delete the following
        return store.data.get("customer", {"email": None})

    def transaction_extras(self, store):
        return store.data.get("extras", None)

    def transaction_invoice(self, store):
        data = store.data["invoice_address"]

        # invoice not requested or address not supplied
        if not data["required"] or not data["address"]:
            return None

        # invoice requested, populate data
        return {
            "address_1": data["address"]["address_1"],
            "address_2": data["address"]["address_2"],
            "address_3": data["address"]["address_3"],
            "address_4": data["address"]["address_4"],
            "address_5": data["address"]["address_5"],
            "country_id": data["address"]["country"]["id"]
            if isinstance(data["address"]["country"], dict)
            else data["address"]["country"],
            "country_state_id": data["address"]["country_state"]["id"]
            if isinstance(data["address"]["country_state"], dict)
            else data["address"]["country_state"],
            "email": data["address"]["email"],
            "phone_1": data["address"]["phone_1"],
            "phone_2": data["address"]["phone_2"],
            "postcode": data["address"]["postcode"],
            "string_1": data["address"]["string_1"],
            "string_2": data["address"]["string_2"],
            "recipient_name": data["address"]["full_name"],
            "recipient_tax_id": data["address"]["tax_id"],
        }

    def transaction_member(self, store):
        if store.user is None:
            return None

        # retrieve user
        user = get_user_model().objects.get(id=store.user)
        if user.salebox_member_id is None:
            return None

        return {
            "salebox_member_id": str(user.salebox_member_id),
            "total_loyalty": store.data["basket"]["loyalty"],
        }

    def transaction_meta(self, store):
        checkout_dt = None
        transaction_dt = [
            store.created.year,
            store.created.month,
            store.created.day,
            store.created.hour,
            store.created.minute,
            store.created.second,
            store.created.microsecond,
        ]

        if store.payment_received is not None:
            checkout_dt = [
                store.created.year,
                store.created.month,
                store.created.day,
                store.created.hour,
                store.created.minute,
                store.created.second,
                store.created.microsecond,
            ]
            transaction_dt = [
                store.payment_received.year,
                store.payment_received.month,
                store.payment_received.day,
                store.payment_received.hour,
                store.payment_received.minute,
                store.payment_received.second,
                store.payment_received.microsecond,
            ]

        return {
            "guid": store.visible_id,
            "sequential_id": self.generate_sequential_id(store),
            "checkout_dt": checkout_dt,
            "user_id": settings.SALEBOX["MISC"]["POS_USER_ID"],
            "utc": transaction_dt,
        }

    def transaction_payment(self, store):
        total_price = (
            store.data["basket"]["sale_price"] + store.data["shipping_method"]["price"]
        )

        return [
            {
                "amount_change": 0,
                "amount_tendered": total_price,
                "method_id": store.data["payment_method"]["id"],
            }
        ]

    def transaction_shipping(self, store):
        address = store.data["shipping_address"]
        method = store.data["shipping_method"]

        # shipping not requested
        if not address["required"]:
            return None

        # shipping requested, populate data
        return {
            "address_1": address["address"]["address_1"],
            "address_2": address["address"]["address_2"],
            "address_3": address["address"]["address_3"],
            "address_4": address["address"]["address_4"],
            "address_5": address["address"]["address_5"],
            "country_id": address["address"]["country"]["id"]
            if isinstance(address["address"]["country"], dict)
            else address["address"]["country"],
            "country_state_id": address["address"]["country_state"]["id"]
            if isinstance(address["address"]["country_state"], dict)
            else address["address"]["country_state"],
            "email": address["address"]["email"],
            "method_id": method["id"],
            "phone_1": address["address"]["phone_1"],
            "phone_2": address["address"]["phone_2"],
            "postcode": address["address"]["postcode"],
            "price": method["price"],
            "string_1": address["address"]["string_1"],
            "string_2": address["address"]["string_2"],
            "recipient_name": address["address"]["full_name"],
        }

    def transaction_total(self, store):
        vat_rate = settings.SALEBOX["MISC"]["VAT_RATE"]

        # calculate gross + gross_vat_applicable
        gross_total = 0
        gross_vat_applicable = 0
        for b in store.data["basket"]["items"]:
            gross_total += b["variant"]["qty_sale_price"]
            if b["variant"]["product"]["vat_applicable"]:
                gross_vat_applicable += b["variant"]["qty_sale_price"]

        # calculate
        total_vat = round(gross_vat_applicable / (1 + (vat_rate / 100)))
        total_vat = gross_vat_applicable - total_vat
        return {
            "total_gross": gross_total,
            "total_net": gross_total - total_vat,
            "total_vat": total_vat,
            "vat_rate": vat_rate,
        }



# Generated by Django 3.2.6 on 2021-08-23 06:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0024_auto_20210821_0002'),
    ]

    operations = [
        migrations.AddField(
            model_name='product',
            name='local_image_version',
            field=models.CharField(blank=True, max_length=70, null=True),
        ),
        migrations.AddField(
            model_name='productcategory',
            name='local_image_version',
            field=models.CharField(blank=True, max_length=70, null=True),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='local_image_version',
            field=models.CharField(blank=True, max_length=70, null=True),
        ),
    ]



import datetime
import requests

from django.conf import settings
from django.contrib.auth import get_user_model
from django.template.loader import render_to_string

try:
    from mailqueue.models import MailerMessage
except:
    pass

from saleboxdjango.lib.common import api_auth
from saleboxdjango.models import Event, Member


class SaleboxEventHandler:
    def __init__(self):
        # get list of events
        es = Event.objects.filter(processed_flag=False).order_by("created")

        # loop through and process them
        for e in es:
            if e.event == "member_created":
                self.event_member_created(e)
            elif e.event == "transaction_created":
                self.event_transaction_created(e)
            elif e.event == "transaction_shipping_packed":
                self.event_transaction_shipping_packed(e)
            elif e.event == "transaction_shipping_picked":
                self.event_transaction_shipping_picked(e)
            elif e.event == "transaction_shipping_shipped":
                self.event_transaction_shipping_shipped(e)

    def event_member_created(self, event):
        """
        Add your code here
        Don't forget to set the processed_flag when done
        """
        event.processed_flag = True
        event.save()
        self._sync_member_history(event)

    def event_transaction_created(self, event):
        try:
            t = self._fetch_transaction(event)

            if t["status"] != "OK":
                return

            # queue email
            email = self._fetch_transaction_email(t)
            if email is not None:
                self._mailqueue(
                    email,
                    render_to_string(
                        "salebox/email/transaction_created/subject.txt",
                        t["transaction"],
                    ),
                    render_to_string(
                        "salebox/email/transaction_created/body.txt", t["transaction"]
                    ),
                    render_to_string(
                        "salebox/email/transaction_created/body.html", t["transaction"]
                    ),
                )

            # mark as processed
            event.processed_flag = True
            event.save()
            self._sync_member_history(event)
        except:
            pass

    def event_transaction_shipping_packed(self, event):
        """
        Add your code here
        Don't forget to set the processed_flag when done
        """
        event.processed_flag = True
        event.save()
        self._sync_member_history(event)

    def event_transaction_shipping_picked(self, event):
        """
        Add your code here
        Don't forget to set the processed_flag when done
        """
        event.processed_flag = True
        event.save()
        self._sync_member_history(event)

    def event_transaction_shipping_shipped(self, event):
        try:
            t = self._fetch_transaction(event)
            if t["status"] != "OK":
                return

            # queue email
            email = self._fetch_transaction_email(t)
            if email is not None:
                self._mailqueue(
                    email,
                    render_to_string(
                        "salebox/email/shipping_shipped/subject.txt", t["transaction"]
                    ),
                    render_to_string(
                        "salebox/email/shipping_shipped/body.txt", t["transaction"]
                    ),
                    render_to_string(
                        "salebox/email/shipping_shipped/body.html", t["transaction"]
                    ),
                )

            # mark as processed
            event.processed_flag = True
            event.save()
            self._sync_member_history(event)
        except:
            pass

    def _fetch_member(self, event):
        member = (
            Member.objects.filter(salebox_member_id=event.salebox_member_id)
            .select_related("parent")
            .first()
        )

        user = (
            get_user_model()
            .objects.filter(salebox_member_id=event.salebox_member_id)
            .first()
        )

        return {"member": member, "user": user}

    def _fetch_transaction(self, event):
        post = api_auth()
        post["pos_guid"] = event.transaction_guid

        # do request
        url = "%s/api/pos/v2/transaction/fetch" % settings.SALEBOX["API"]["URL"]
        try:
            r = requests.post(url, data=post)
            o = r.json()
            o["transaction"]["dt"] = datetime.datetime(*o["transaction"]["dt"])
            return o
        except:
            return None

    def _fetch_transaction_email(self, transaction):
        if transaction["transaction"]["customer_email"] is not None:
            return transaction["transaction"]["customer_email"]

        if (
            transaction["transaction"]["member"] is not None
            and "email" in transaction["transaction"]["member"]
        ):
            return transaction["transaction"]["member"]["email"]

        return None

    def _sync_member_history(self, event):
        if event.salebox_member_id is not None:
            m = Member.objects.filter(salebox_member_id=event.salebox_member_id).first()
            if m is not None:
                m.transactionhistory_request_sync()

    # optional: for use with django-mail-queue
    def _mailqueue(
        self,
        to_address,
        subject,
        content,
        html=None,
        cc=None,
        bcc=None,
        from_address=None,
    ):
        try:
            MailerMessage
        except:
            return

        # from address
        if from_address is None:
            from_address = settings.DEFAULT_FROM_EMAIL

        # create message
        msg = MailerMessage()
        msg.from_address = from_address
        msg.subject = subject.replace("\n", "").strip()
        msg.to_address = to_address
        msg.content = content
        msg.app = "Salebox"

        # optional extras
        if html is not None:
            msg.html_content = html
        if cc is not None:
            msg.cc_address = cc
        if bcc is not None:
            msg.bcc_address = bcc

        # save in queue
        msg.save()



# Generated by Django 3.0.4 on 2020-03-09 05:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0016_auto_20191216_1359'),
    ]

    operations = [
        migrations.AlterField(
            model_name='productcategory',
            name='level',
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='lft',
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='rght',
            field=models.PositiveIntegerField(editable=False),
        ),
    ]



from django.urls import path

# address
from .views.address.add import SaleboxAddressAddView
from .views.address.remove import SaleboxAddressRemoveView
from .views.address.set_default import SaleboxAddressSetDefaultView

# analytics
from .views.analytics.analytics import SaleboxAnalyticsView

# basket
from .views.basket.basket import SaleboxBasketBasketView
from .views.basket.migrate import SaleboxBasketMigrateView
from .views.basket.wishlist import SaleboxBasketWishlistView

# image
from saleboxdjango.views.image.image import salebox_image_view

# rating
from .views.rating.add import SaleboxRatingAddView
from .views.rating.remove import SaleboxRatingRemoveView

# sync
from .views.sync.sync import SaleboxSyncView

urlpatterns = [
    # address
    path("address/add/", SaleboxAddressAddView.as_view()),
    path("address/remove/", SaleboxAddressRemoveView.as_view()),
    path("address/set-default/", SaleboxAddressSetDefaultView.as_view()),
    # analytics
    path("analytics/", SaleboxAnalyticsView.as_view()),
    # basket
    path("basket/basket/", SaleboxBasketBasketView.as_view()),
    path("basket/migrate/", SaleboxBasketMigrateView.as_view()),
    path("basket/wishlist/", SaleboxBasketWishlistView.as_view()),
    # image
    path(
        "img/<slug:imgtype>/<slug:dir>/<int:id>.<slug:hash>.<slug:suffix>",
        salebox_image_view,
    ),
    # rating
    path("rating/add/", SaleboxRatingAddView.as_view()),
    path("rating/remove/", SaleboxRatingRemoveView.as_view()),
    # sync
    path("sync/", SaleboxSyncView),
]



from saleboxdjango.views.basket.base import SaleboxBasketView
from saleboxdjango.forms import SaleboxBasketBasketForm


class SaleboxBasketBasketView(SaleboxBasketView):
    action = "basket-basket"
    form = SaleboxBasketBasketForm

    def form_valid(self, request):
        try:
            self.sb.update_basket(
                request,
                self.form.cleaned_data["variant_id"],
                self.form.cleaned_data["quantity"],
                self.form.cleaned_data["relative"],
            )
            self.status = "success"

            self.results = self.sb.get_data(
                request, self.results_csv, self.form.cleaned_data["variant_id"]
            )
        except:
            self.status = "fail"



# Generated by Django 2.1.8 on 2019-12-16 06:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0013_analytic'),
    ]

    operations = [
        migrations.AddField(
            model_name='analytic',
            name='language',
            field=models.CharField(blank=True, max_length=10, null=True),
        ),
        migrations.AddField(
            model_name='analytic',
            name='screen_height',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='analytic',
            name='screen_width',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_browser_family',
            field=models.CharField(blank=True, max_length=32, null=True, verbose_name='Browser family'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_browser_version',
            field=models.CharField(blank=True, max_length=20, null=True, verbose_name='Browser version'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_device_brand',
            field=models.CharField(blank=True, max_length=20, null=True, verbose_name='Device brand'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_device_family',
            field=models.CharField(blank=True, max_length=32, null=True, verbose_name='Device family'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_device_model',
            field=models.CharField(blank=True, max_length=20, null=True, verbose_name='Device model'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_is_bot',
            field=models.BooleanField(null=True, verbose_name='Is bot?'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_is_mobile',
            field=models.BooleanField(null=True, verbose_name='Is mobile?'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_is_pc',
            field=models.BooleanField(null=True, verbose_name='Is PC?'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_is_tablet',
            field=models.BooleanField(null=True, verbose_name='Is tablet?'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_is_touch_capable',
            field=models.BooleanField(null=True, verbose_name='Is touchscreen?'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_os_family',
            field=models.CharField(blank=True, max_length=20, null=True, verbose_name='OS family'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ua_os_version',
            field=models.CharField(blank=True, max_length=20, null=True, verbose_name='OS version'),
        ),
    ]



from datetime import datetime
from django.core.management.base import BaseCommand

from saleboxdjango.models import KeyValueStore
from saleboxdjango.management.commands.lib.log import log
from saleboxdjango.management.commands.lib.sync_pull_queue import SaleboxPullQueue
from saleboxdjango.management.commands.lib.sync_pull_models import SaleboxPullModels
from saleboxdjango.management.commands.lib.sync_pull_posimages import (
    SaleboxPullPOSImages,
)
from saleboxdjango.management.commands.lib.sync_pull_inventory import (
    SaleboxPullInventory,
)
from saleboxdjango.management.commands.lib.sync_push_footfall import SaleboxPushFootfall
from saleboxdjango.models import CheckoutStore


class Command(BaseCommand):
    def handle(self, *args, **options):
        if self.get_sync_in_progress():
            print("Bailing out: sync is in progress")
            return

        # start
        log("core", "saleboxsync2 starting")
        self.set_sync_in_progress(True)

        # do stuff
        SaleboxPullQueue()
        SaleboxPullModels()
        SaleboxPullPOSImages()
        SaleboxPullInventory()
        SaleboxPushFootfall()

        # clean up checkout store
        CheckoutStore.objects.apply_timeout()

        # done
        log("core", "saleboxsync2 completed successfully")
        self.set_sync_in_progress(False)

    def get_sync_in_progress(self):
        kvs = KeyValueStore.objects.filter(key="sync_in_progress").first()

        # first time here? create the keyvalue record
        if kvs is None:
            self.set_sync_in_progress(False)
            return False

        # if sync is in progress, did it start long enough ago to
        # assume it failed?
        if kvs.value["sync_in_progress"]:
            diff = datetime.utcnow().timestamp() - kvs.value["last_seen"]
            if diff > 300:  # 5 minutes
                log("core", "saleboxsync2 sync_in_progress timeout", "WARNING")
                return False

        # fallback
        return kvs.value["sync_in_progress"]

    def set_sync_in_progress(self, sync_in_progress):
        # create new value
        value = {
            "sync_in_progress": sync_in_progress,
            "last_seen": datetime.utcnow().timestamp(),
        }

        # retrieve from the db
        kvs = KeyValueStore.objects.filter(key="sync_in_progress").first()
        if kvs is None:
            KeyValueStore(key="sync_in_progress", value=value).save()
        else:
            kvs.value = value
            kvs.save()



# Generated by Django 3.2.6 on 2021-08-14 11:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0021_auto_20210814_1825'),
    ]

    operations = [
        migrations.AddField(
            model_name='attribute',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='attribute',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='attributeitem',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='attributeitem',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='contentpage',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='contentpage',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='contentpageitem',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='contentpageitem',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='country',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='country',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='countrystate',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='countrystate',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='discountgroup',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='discountgroup',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='discountruleset',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='discountruleset',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='member',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='member',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='membergroup',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='membergroup',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='product',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productcategory',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productcategory',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productvariant',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productvariantimage',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='productvariantimage',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='translation',
            name='created',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='translation',
            name='last_update',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]



# Generated by Django 2.1.8 on 2019-12-16 06:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0015_auto_20191216_1357'),
    ]

    operations = [
        migrations.AlterField(
            model_name='analytic',
            name='ip_city',
            field=models.CharField(blank=True, max_length=32, null=True, verbose_name='City'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ip_country',
            field=models.CharField(blank=True, max_length=5, null=True, verbose_name='Country'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ip_lat',
            field=models.DecimalField(blank=True, decimal_places=6, max_digits=9, null=True, verbose_name='Latitude'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ip_lng',
            field=models.DecimalField(blank=True, decimal_places=6, max_digits=9, null=True, verbose_name='Longitude'),
        ),
        migrations.AlterField(
            model_name='analytic',
            name='ip_tz',
            field=models.CharField(blank=True, max_length=32, null=True, verbose_name='Timezone'),
        ),
    ]



from django.http import HttpResponse, HttpResponseRedirect
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic import View

from saleboxdjango.models import CallbackStore, CheckoutStore, CheckoutStoreUpdate


@method_decorator(csrf_exempt, name="dispatch")
class SaleboxCallbackView(View):
    get_redirect = None
    post_redirect = None

    def get_ip(self, request):
        x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
        if x_forwarded_for:
            ip = x_forwarded_for.split(",")[0]
        else:
            ip = request.META.get("REMOTE_ADDR")
        return ip

    def dispatch(self, request, *args, **kwargs):
        CallbackStore(
            ip_address=self.get_ip(request),
            method=request.method.lower(),
            post=request.POST,
        ).save()

        return super().dispatch(request, *args, **kwargs)

    def get(self, request):
        if self.get_redirect is not None:
            return HttpResponseRedirect(self.get_redirect)
        return HttpResponse("OK")

    def post(self, request):
        if self.post_redirect is not None:
            return HttpResponseRedirect(self.post_redirect)
        return HttpResponse("OK")

    def _get_store(self, uuid, visible_id):
        return (
            CheckoutStore.objects.filter(uuid=uuid)
            .filter(visible_id=visible_id)
            .first()
        )

    def _save_store_update(self, store, status, data):
        CheckoutStoreUpdate(
            store=store,
            status=status,
            data=data,
        ).save()



# -*- coding: utf-8 -*-
from distutils.core import setup
from setuptools import find_packages

setup(
    name="salebox-django",
    version="0.0.247",
    author="Jon Combe",
    author_email="jon@salebox.io",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "django-mptt",
        "django-opposable-thumbs",
        "geoip2",
        "pyShipping-python3",
        "pyyaml",
        "requests",
        "ua-parser",
        "user-agents",
    ],
    url="https://salebox.io",
    license="BSD licence, see LICENCE file",
    description="Django ecommerce powered by Salebox",
    long_description="Connect a standalone Django ecommerce website to a salebox.io backoffice",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Environment :: Web Environment",
        "Framework :: Django",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
    zip_safe=False,
)



# Generated by Django 3.2.12 on 2024-03-14 06:03

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0027_auto_20220516_0933'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='analytic',
            options={'ordering': ['-last_seen']},
        ),
    ]



import json

from django.conf import settings
from django.http import JsonResponse
from django.template.loader import render_to_string
from django.utils.crypto import get_random_string
from django.utils.translation import get_language

from saleboxdjango.lib.translation import get_translation, get_translations
from saleboxdjango.models import Country, CountryState, UserAddress


class SaleboxAddress:
    def __init__(self, user, address_group="default"):
        self.user = user
        if self.user.is_authenticated:
            self.query = (
                UserAddress.objects.filter(user=user)
                .filter(address_group=address_group)
                .filter(active_flag=True)
                .select_related("country", "country_state")
            )

    def add(self, values):
        # get country
        if isinstance(values["country"], int):
            values["country"] = Country.objects.get(id=values["country"])

        # get country_state
        if isinstance(values["country_state"], int):
            values["country_state"] = CountryState.objects.get(
                id=values["country_state"]
            )

        # create address instance
        address = UserAddress(
            default=values["default"] or False,
            address_group=values["address_group"] or "default",
            full_name=values["full_name"],
            address_1=values["address_1"],
            address_2=values["address_2"],
            address_3=values["address_3"],
            address_4=values["address_4"],
            address_5=values["address_5"],
            country_state=values["country_state"],
            country=values["country"],
            postcode=str(values["postcode"] or "").upper(),
            phone_1=values["phone_1"],
            phone_2=values["phone_2"],
            email=values["email"],
            string_1=values["string_1"],
            string_2=values["string_2"],
            tax_id=values["tax_id"],
        )

        # save to db if user is authenticated
        if self.user.is_authenticated:
            address.user = self.user
            address.save()

        return address

    def form_extras(self, country_id=None):
        country_list = self._get_country_list()
        country_state_lookup = self._get_country_state_lookup()

        # build 'current' country_state list
        try:
            current_country_states = country_state_lookup[int(country_id)]
        except:
            current_country_states = []

        # build javascript country_state dict
        js_country_states = (
            json.dumps(country_state_lookup, ensure_ascii=False)
            .replace('"id": ', "i:")
            .replace(', "name": ', ",s:")
            .replace("}, {", "},{")
        )

        return {
            "country_list": country_list,
            "current_states": current_country_states,
            "js_states": js_country_states,
        }

    def get(self, selected_id=None, force_selected=False, non_null_fields=[]):
        # return an empty list for non-authenticated users
        if not self.user.is_authenticated:
            return []

        # get addresses
        addresses = self.query.all()
        if "tax_id" in non_null_fields:
            addresses = addresses.filter(tax_id__isnull=False)

        # add 'selected' value and localised country name
        for a in addresses:
            a.selected = False
            # a.country_name = self._get_country_name(a.country)
            # a.country_state_name = self._get_country_state_name(a.country_state)

        # set selected_id if exists
        has_selected = False
        for a in addresses:
            if a.id == selected_id:
                a.selected = True
                has_selected = True

        # force selection of an address
        if force_selected and not has_selected and len(addresses) > 0:
            # select the default address
            for a in addresses:
                if a.default:
                    a.selected = True
                    has_selected = True
                    break

            # fall back to the first address
            if not has_selected:
                addresses[0].selected = True

        # return
        return addresses

    def get_count(self):
        return self.query.all().count()

    def set_default(self, address_id):
        address = self._get_address_by_id(address_id)
        if address is not None and not address.default:
            address.default = True
            address.save()

    def set_inactive(self, address_id):
        address = self._get_address_by_id(address_id)
        if address is not None:
            address.active_flag = False
            address.save()

    def _get_address_by_id(self, address_id):
        return self.query.all().filter(id=address_id).first()

    def _get_allowed_countries(self):
        try:
            return settings.SALEBOX["MISC"]["ALLOWED_COUNTRIES"]
        except:
            return []

    def _get_country_list(self):
        lang = get_language()

        countries = Country.objects.all()
        allowed_countries = self._get_allowed_countries()
        if len(allowed_countries) > 0:
            countries = countries.filter(id__in=allowed_countries)
        countries = list(countries.values("id", "code"))

        # get translated name
        for c in countries:
            c["name"] = get_translation(lang, "place.%s" % c["code"], c["code"])
            del c["code"]

        # sort
        countries = sorted(countries, key=lambda k: k["name"])
        return countries

    def _get_country_state_lookup(self):
        lang = get_language()
        i18n = get_translations()

        states_list = CountryState.objects.all()
        allowed_countries = self._get_allowed_countries()
        if len(allowed_countries) > 0:
            states_list = states_list.filter(country__id__in=allowed_countries)
        states_list = list(states_list.values("id", "country__id", "full_code"))

        # create a lookup
        lookup = {}
        for s in states_list:
            lookup[s["id"]] = {
                "country": s["country__id"],
                "name": get_translation(
                    lang, "place.%s" % s["full_code"], translations=i18n
                ),
            }

        # create dict of states per country
        countries = {}
        for key in lookup:
            if lookup[key]["country"] not in countries:
                countries[lookup[key]["country"]] = []
            countries[lookup[key]["country"]].append(
                {"id": key, "name": lookup[key]["name"]}
            )

        # sort each country
        if not lang.startswith("en"):
            for c in countries:
                countries[c] = sorted(countries[c], key=lambda k: k["name"])

        return countries



from os import makedirs
import requests

from django.conf import settings
from django.db.models import F

from saleboxdjango.models import ProductCategory, \
    Product, ProductVariant


class SaleboxPullPOSImages:
    def __init__(self):
        self.update_productcategories()
        self.update_products()
        self.update_productvariants()

    def get_folder(self, dir):
        target_dir = f'{settings.MEDIA_ROOT}/salebox/{dir}'
        try:
            makedirs(target_dir)
        except:
            pass

        return target_dir

    def sync_image(self, id, urlpath, dir):
        target_dir = self.get_folder(dir)

        # fetch image
        try:
            url = f'{settings.SALEBOX["API"]["URL"]}/{urlpath}'
            suffix = urlpath.split('.')[-1]
            filename = f'{id}.{urlpath.split("/")[-1][0:6]}.{suffix}'
            target = f'{target_dir}/{filename}'

            r = requests.get(url)
            if r.status_code == 200:
                open(target, 'wb').write(r.content)
                return filename, True
        except:
            print('ERROR')

        return None, False

    def update_products(self):
        products = Product \
                    .objects \
                    .exclude(image__isnull=True) \
                    .exclude(local_image_version__exact=F('image'))

        for p in products:
            path, success = self.sync_image(p.id, p.image[1:], 'posp')
            if success:
                p.local_image_version = p.image
                p.local_image = path
                p.save()

    def update_productcategories(self):
        categories = ProductCategory \
                        .objects \
                        .exclude(image__isnull=True) \
                        .exclude(local_image_version__exact=F('image'))

        for c in categories:
            path, success = self.sync_image(c.id, c.image[1:], 'pospc')
            if success:
                c.local_image_version = c.image
                c.local_image = path
                c.save()

    def update_productvariants(self):
        variants = ProductVariant \
                    .objects \
                    .exclude(image__isnull=True) \
                    .exclude(local_image_version__exact=F('image'))

        for v in variants:
            path, success = self.sync_image(v.id, v.image[1:], 'pospv')
            if success:
                v.local_image_version = v.image
                v.local_image = path
                v.save()


from django import template
from django.utils.translation import get_language

from saleboxdjango.lib.translation import get_translation
from saleboxdjango.models import Country, CountryState

register = template.Library()


@register.simple_tag
def sb_country_name(country, default="", lang=None):
    if isinstance(country, dict):
        country = country["id"]
    if isinstance(country, int):
        country = Country.objects.get(id=country)
    if not isinstance(country, Country):
        return default

    if lang is None:
        lang = (get_language()).lower().split("-")[0]

    return get_translation(lang, "place.%s" % country.code, default)


@register.simple_tag
def sb_country_state_name(country_state, default="", lang=None):
    if isinstance(country_state, dict):
        country_state = country_state["id"]
    if isinstance(country_state, int):
        country_state = CountryState.objects.get(id=country_state)
    if not isinstance(country_state, CountryState):
        return default

    if lang is None:
        lang = (get_language()).lower().split("-")[0]

    return get_translation(lang, "place.%s" % country_state.full_code, default)



from django.shortcuts import redirect

from django.views.generic import TemplateView

from saleboxdjango.lib.category import SaleboxCategory
from saleboxdjango.lib.product import SaleboxProduct
from saleboxdjango.models import ProductCategory, ProductVariant


class SaleboxProductsView(TemplateView):
    page_number = 1
    path = []

    def dispatch(self, request, *args, **kwargs):
        # retrieve our url path
        self.path = self.kwargs["path"].strip("/").split("/")
        self.path = [p.strip() for p in self.path if len(p.strip()) > 0]

        # retrieve page number
        if len(self.path) > 0:
            if self._is_positive_integer(self.path[-1]):
                self.page_number = int(self.path[-1])
                self.path.pop()

        return super().dispatch(request, *args, **kwargs)

    def get(self, request, *args, **kwargs):
        if len(self.path) == 0:
            return self.product_list(None, request, *args, **kwargs)

        # is this a product?
        if len(self.path) == 2 and self.path[0] == "product":
            id = int(self.path[-1].split("-")[0])
            pv = ProductVariant.objects.filter(active_flag=True).filter(id=id).first()
            return self.product_detail(pv, request, *args, **kwargs)

        # this is a category
        pc = (
            ProductCategory.objects.filter(active_flag=True)
            .filter(slug=self.path[-1])
            .first()
        )
        if pc:
            return self.product_list(pc, request, *args, **kwargs)

        # go up a level
        return redirect("/")

    def product_detail(self, variant, request, *args, **kwargs):
        self.template_name = "salebox/product/product_detail.html"

        #
        context = self.get_context_data(**kwargs)
        context["variant"] = variant

        # render
        return self.render_to_response(context)

    def product_list(self, category, request, *args, **kwargs):
        self.template_name = "salebox/product/product_list.html"
        context = self.get_context_data(**kwargs)

        # fetch categories
        sc = SaleboxCategory()
        context["categories"] = sc.get_tree(
            cache_key="product_categories",
            category_id=category.id if category else None,
        )

        # fetch products
        sp = SaleboxProduct()
        pagination_suffix = ""
        if len(self.path) > 0:
            pagination_suffix = "%s/" % "/".join(self.path)
        sp.set_pagination(self.page_number, 20, "/en/shop/%s" % pagination_suffix)
        if category:
            sp.set_category(category)
        context["products"] = sp.get_list()

        # render
        return self.render_to_response(context)

    def _is_positive_integer(self, value):
        try:
            i = int(value)
            if i > 0:
                return True
        except:
            pass

        return False



from saleboxdjango.lib.binpack import BinPack


class SaleboxShippingOptions:
    DEFAULT_SELECT = "FIRST"  # or 'CHEAPEST' or None
    REMOVE_UNAVAILABLE = True
    SORT_BY = "PRICE_ASC"  # 'LABEL' or None

    # variant variable names
    SHIPPING_WIDTH = "int_1"
    SHIPPING_HEIGHT = "int_2"
    SHIPPING_DEPTH = "int_3"
    SHIPPING_WEIGHT = "shipping_weight"

    def get_options(self):
        # This is the function you need to replace.
        # Add your own functions here, e.g.
        #
        #   return [
        #     self._example_option_1(),
        #     self._example_option_2(),
        #   ]
        #
        # Useful vars:
        #   self.checkout['shipping_address']['address']['country']
        #   self.checkout['shipping_address']['address']['country_state']
        #
        return [
            self._example_option_1(),
            self._example_option_2(),
        ]

    def go(self, request, checkout, context):
        self.request = request
        self.checkout = checkout
        self.context = context

        # get packages into a handy list for bin packing
        self.meta = {
            "total_items": 0,
            "total_items_onshelf": 0,
            "total_items_preorder": 0,
            "total_weight": 0,
            "total_weight_onshelf": 0,
            "total_weight_preorder": 0,
            "items": [],
            "items_onshelf": [],
            "items_preorder": [],
        }
        for b in checkout["basket"]["items"]:
            for i in range(0, b["qty"]):
                self.meta["total_items"] += 1
                self.meta["total_weight"] += b["variant"][self.SHIPPING_WEIGHT] or 0
                self.meta["items"].append(
                    {
                        "variant_id": b["variant"]["id"],
                        "width": b["variant"][self.SHIPPING_WIDTH],
                        "height": b["variant"][self.SHIPPING_HEIGHT],
                        "depth": b["variant"][self.SHIPPING_DEPTH],
                        "weight": b["variant"][self.SHIPPING_WEIGHT],
                    }
                )

                if b["variant"]["preorder_flag"]:
                    self.meta["total_items_preorder"] += 1
                    self.meta["total_weight_preorder"] += (
                        b["variant"][self.SHIPPING_WEIGHT] or 0
                    )
                    self.meta["items_preorder"].append(
                        {
                            "variant_id": b["variant"]["id"],
                            "width": b["variant"][self.SHIPPING_WIDTH],
                            "height": b["variant"][self.SHIPPING_HEIGHT],
                            "depth": b["variant"][self.SHIPPING_DEPTH],
                            "weight": b["variant"][self.SHIPPING_WEIGHT],
                        }
                    )
                else:
                    self.meta["total_items_onshelf"] += 1
                    self.meta["total_weight_onshelf"] += (
                        b["variant"][self.SHIPPING_WEIGHT] or 0
                    )
                    self.meta["items_onshelf"].append(
                        {
                            "variant_id": b["variant"]["id"],
                            "width": b["variant"][self.SHIPPING_WIDTH],
                            "height": b["variant"][self.SHIPPING_HEIGHT],
                            "depth": b["variant"][self.SHIPPING_DEPTH],
                            "weight": b["variant"][self.SHIPPING_WEIGHT],
                        }
                    )

        # build options
        opts = self.get_options()

        # remove nulls
        # if an shipping option is not available, simply return None. However...
        opts = [o for o in opts if o is not None]

        # optional: remove unavailable
        # you *may* want to keep unavailable options in the list but greyed out,
        # in which case, return the dict with ['available'] = False
        if self.REMOVE_UNAVAILABLE:
            opts = [o for o in opts if o["available"] == True]

        # set selected value (if exists)
        if self.checkout["shipping_method"]["id"] is not None:
            option = next(
                o for o in opts if o["id"] == self.checkout["shipping_method"]["id"]
            )
            option["selected"] = True

        # optional: sorting
        if self.SORT_BY == "PRICE_ASC":
            opts = sorted(opts, key=lambda k: k["price"])
        elif self.SORT_BY == "LABEL":
            opts = sorted(opts, key=lambda k: k["label"]["label"])

        # optional: default select
        if len(opts) > 0:
            if self.DEFAULT_SELECT == "CHEAPEST":
                tmp = sorted(opts, key=lambda k: k["price"])
                id = tmp[0]["id"]
                option = next(o for o in opts if o["id"] == id)
                option["selected"] = True
            elif self.DEFAULT_SELECT == "FIRST":
                opts[0]["selected"] = True

        # calculate combined_price
        for o in opts:
            o["combined_price"] = self.checkout["basket"]["sale_price"] + o["price"]

        # return
        self.context["shipping_options"] = opts
        return self.context

    def init_option(self, id, code, label, remarks, service, meta=None):
        return {
            "available": True,
            "code": code,
            "id": id,
            "label": {
                "label": label,  # e.g. 'Post Office'
                "remarks": remarks,  # e.g. '2-3 days'
                "service": service,  # 'ExpressPost'
            },
            "meta": meta,
            "price": 0,
            "selected": False,
        }

    def _do_binpack(self, containers, packages):
        # containers example:
        # [
        #   ['name', price, width, height, depth],
        #   ['name', price, width, height, depth],
        # ]
        bp = BinPack()

        # add containers
        for c in containers:
            bp.add_bin(c[0], c[1], c[2], c[3], c[4])

        # add packages
        for p in packages:
            bp.add_package(
                p["variant_id"],
                p["width"] or 0,
                p["height"] or 0,
                p["depth"] or 0,
            )

        return bp.go()

    def _example_option_1(self):
        method = self.init_option(
            1, "postoffice", "Post Office", "2-4 days", "Surface mail"
        )

        method["price"] = 10000
        return method

    def _example_option_2(self):
        method = self.init_option(
            2, "postofficeexpress", "Post Office", "1 day", "NextDay"
        )

        method["price"] = 25000
        return method



# Generated by Django 3.2.6 on 2021-12-31 06:42

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0025_auto_20210823_1313'),
    ]

    operations = [
        migrations.AlterField(
            model_name='analytic',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='attribute',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='attributeitem',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='basketwishlist',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='callbackstore',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='callbackstore',
            name='post',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='checkoutstore',
            name='data',
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name='checkoutstore',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='checkoutstoreupdate',
            name='data',
            field=models.JSONField(),
        ),
        migrations.AlterField(
            model_name='checkoutstoreupdate',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='contentkeyvaluestore',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='contentpage',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='contentpageitem',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='country',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='countrystate',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='discountgroup',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='discountruleset',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='event',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='keyvaluestore',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='keyvaluestore',
            name='value',
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='lastupdate',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='member',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='member',
            name='salebox_transactionhistory_data',
            field=models.JSONField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='membergroup',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='product',
            name='i18n',
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='product',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='i18n',
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='productcategory',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productvariant',
            name='i18n',
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name='productvariant',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productvariantimage',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='productvariantrating',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='synclog',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='syncqueue',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='translation',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
        migrations.AlterField(
            model_name='useraddress',
            name='id',
            field=models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]



# Generated by Django 3.2.12 on 2022-05-16 02:33

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('saleboxdjango', '0026_auto_20211231_1342'),
    ]

    operations = [
        migrations.AddField(
            model_name='attributeitem',
            name='extras',
            field=models.JSONField(blank=True, db_index=True, default=dict),
        ),
        migrations.AlterField(
            model_name='attributeitem',
            name='slug',
            field=models.CharField(blank=True, db_index=True, max_length=255),
        ),
        migrations.AlterField(
            model_name='attributeitem',
            name='value',
            field=models.CharField(max_length=255),
        ),
    ]
